/****************************************************************
 * Licensed to the Apache Software Foundation (ASF) under one   *
 * or more contributor license agreements.  See the NOTICE file *
 * distributed with this work for additional information        *
 * regarding copyright ownership.  The ASF licenses this file   *
 * to you under the Apache License, Version 2.0 (the            *
 * "License"); you may not use this file except in compliance   *
 * with the License.  You may obtain a copy of the License at   *
 *                                                              *
 *   http://www.apache.org/licenses/LICENSE-2.0                 *
 *                                                              *
 * Unless required by applicable law or agreed to in writing,   *
 * software distributed under the License is distributed on an  *
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY       *
 * KIND, either express or implied.  See the License for the    *
 * specific language governing permissions and limitations      *
 * under the License.                                           *
 ****************************************************************/

package org.apache.james.webadmin.service;

import java.util.Optional;
import java.util.concurrent.atomic.AtomicLong;
import java.util.function.Function;

import javax.mail.MessagingException;

import org.apache.james.json.DTOModule;
import org.apache.james.mailrepository.api.MailKey;
import org.apache.james.mailrepository.api.MailRepositoryPath;
import org.apache.james.mailrepository.api.MailRepositoryStore;
import org.apache.james.server.task.json.dto.TaskDTO;
import org.apache.james.server.task.json.dto.TaskDTOModule;
import org.apache.james.task.Task;
import org.apache.james.task.TaskExecutionDetails;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;

public class ReprocessingAllMailsTask implements Task {

    public static final String TYPE = "reprocessingAllTask";

    public static class AdditionalInformation implements TaskExecutionDetails.AdditionalInformation {
        private final MailRepositoryPath repositoryPath;
        private final String targetQueue;
        private final Optional<String> targetProcessor;
        private final long initialCount;
        private final AtomicLong processedCount;

        public AdditionalInformation(MailRepositoryPath repositoryPath, String targetQueue, Optional<String> targetProcessor, long initialCount) {
            this.repositoryPath = repositoryPath;
            this.targetQueue = targetQueue;
            this.targetProcessor = targetProcessor;
            this.initialCount = initialCount;
            this.processedCount = new AtomicLong(0);
        }

        public String getTargetQueue() {
            return targetQueue;
        }

        public Optional<String> getTargetProcessor() {
            return targetProcessor;
        }

        public String getRepositoryPath() {
            return repositoryPath.asString();
        }

        public long getRemainingCount() {
            return initialCount - processedCount.get();
        }

        public long getInitialCount() {
            return initialCount;
        }

        @JsonIgnore
        public void notifyProgress(MailKey key) {
            processedCount.incrementAndGet();
        }
    }

    public static class UrlEncodingFailureSerializationException extends RuntimeException {

        public UrlEncodingFailureSerializationException(MailRepositoryPath mailRepositoryPath) {
            super("Unable to serialize: '" + mailRepositoryPath + "' can not be url encoded");
        }
    }

    public static class InvalidMailRepositoryPathDeserializationException extends RuntimeException {

        public InvalidMailRepositoryPathDeserializationException(String mailRepositoryPath) {
            super("Unable to deserialize: '" + mailRepositoryPath + "' can not be url decoded");
        }
    }

    private static class ReprocessingAllMailsTaskDTO implements TaskDTO {

        public static ReprocessingAllMailsTaskDTO toDTO(ReprocessingAllMailsTask domainObject, String typeName) {
            try {
                return new ReprocessingAllMailsTaskDTO(
                    typeName,
                    domainObject.additionalInformation.initialCount,
                    domainObject.repositoryPath.urlEncoded(),
                    domainObject.targetQueue,
                    domainObject.targetProcessor
                );
            } catch (Exception e) {
                throw new UrlEncodingFailureSerializationException(domainObject.repositoryPath);
            }
        }

        private final String type;
        private final long repositorySize;
        private final String repositoryPath;
        private final String targetQueue;
        private final Optional<String> targetProcessor;

        public ReprocessingAllMailsTaskDTO(@JsonProperty("type") String type,
                                           @JsonProperty("repositorySize") long repositorySize,
                                           @JsonProperty("repositoryPath") String repositoryPath,
                                           @JsonProperty("targetQueue") String targetQueue,
                                           @JsonProperty("targetProcessor") Optional<String> targetProcessor) {
            this.type = type;
            this.repositorySize = repositorySize;
            this.repositoryPath = repositoryPath;
            this.targetQueue = targetQueue;
            this.targetProcessor = targetProcessor;
        }

        public ReprocessingAllMailsTask fromDTO(ReprocessingService reprocessingService) {
            try {
                return new ReprocessingAllMailsTask(
                    reprocessingService,
                    repositorySize,
                    MailRepositoryPath.fromEncoded(repositoryPath),
                    targetQueue,
                    targetProcessor
                );
            } catch (Exception e) {
                throw new InvalidMailRepositoryPathDeserializationException(repositoryPath);
            }
        }

        @Override
        public String getType() {
            return type;
        }

        public long getRepositorySize() {
            return repositorySize;
        }

        public String getRepositoryPath() {
            return repositoryPath;
        }

        public String getTargetQueue() {
            return targetQueue;
        }

        public Optional<String> getTargetProcessor() {
            return targetProcessor;
        }
    }

    public static final Function<ReprocessingService, TaskDTOModule<ReprocessingAllMailsTask, ReprocessingAllMailsTaskDTO>> MODULE = (reprocessingService) ->
        DTOModule
            .forDomainObject(ReprocessingAllMailsTask.class)
            .convertToDTO(ReprocessingAllMailsTaskDTO.class)
            .toDomainObjectConverter(dto -> dto.fromDTO(reprocessingService))
            .toDTOConverter(ReprocessingAllMailsTaskDTO::toDTO)
            .typeName(TYPE)
            .withFactory(TaskDTOModule::new);

    private final ReprocessingService reprocessingService;
    private final MailRepositoryPath repositoryPath;
    private final String targetQueue;
    private final Optional<String> targetProcessor;
    private final AdditionalInformation additionalInformation;

    public ReprocessingAllMailsTask(ReprocessingService reprocessingService, long repositorySize,
                                    MailRepositoryPath repositoryPath, String targetQueue, Optional<String> targetProcessor) {
        this.reprocessingService = reprocessingService;
        this.repositoryPath = repositoryPath;
        this.targetQueue = targetQueue;
        this.targetProcessor = targetProcessor;
        this.additionalInformation = new AdditionalInformation(
            repositoryPath, targetQueue, targetProcessor, repositorySize);
    }

    @Override
    public Result run() {
        try {
            reprocessingService.reprocessAll(repositoryPath, targetProcessor, targetQueue, additionalInformation::notifyProgress);
            return Result.COMPLETED;
        } catch (MessagingException | MailRepositoryStore.MailRepositoryStoreException e) {
            LOGGER.error("Encountered error while reprocessing repository", e);
            return Result.PARTIAL;
        }
    }

    @Override
    public String type() {
        return TYPE;
    }

    @Override
    public Optional<TaskExecutionDetails.AdditionalInformation> details() {
        return Optional.of(additionalInformation);
    }

}
